from threading import Lock
from threadsnake.myhttp.logging import LogColors
from threading import Thread
import time
from typing import Any, List, Dict
from random import random

class SharpLockLocks:
    locks:Dict[int, Lock] = dict()  

class SharpLock(SharpLockLocks):
    def __init__(self, obj:Any):
        self.lockid = id(obj)
        if self.lockid not in SharpLockLocks.locks:
            SharpLockLocks.locks[self.lockid] = Lock()
    
    def __enter__(self):
        SharpLockLocks.locks[self.lockid].acquire()
        return self
    
    def __exit__(self, type, value, traceback):
        SharpLockLocks.locks[self.lockid].release()
        

class Progress(Thread):
    def __init__(self, length:int=25, value:int=0, interval:float=0.1, emptyCharacter:str = '.', fillCharacter:str = '|'):
        Thread.__init__(self)
        
        self.running:bool = False
        self.length:int = length
        self.value:float = value
        self.interval:float = interval
        
        self.textColor:str = LogColors.BRIGHT_FOREGROUND_WHITE
        self.barColors:List[str] = [LogColors.BRIGHT_FOREGROUND_GREEN]
        self.percentColor:str = [LogColors.BRIGHT_FOREGROUND_YELLOW, LogColors.BRIGHT_FOREGROUND_GREEN]
        
        self.animationCharacters:List[str] = [' ']
        self.animationIndex = 0
        self.size = 10
        self.emptyBarCharacter = ' '
        self.fillBarCharacter = '█'
        
        self.show:bool = True
        
    def set_message(self, text:str):
        text = ' '*self.size + text + ' '*self.size
        self.animationCharacters:List[str] = [text[:self.size] for i in range(10)] 
        self.animationCharacters += [text[i:i+self.size] for i in range(len(text)-self.size)]
    
    def set_value(self, value:float = 0):
        self.show = True
        self.value = min(max(0, value), 1)
    
    def carriage_return(self):
        print('\r', end='')
    
    def animate_text(self):
        if len(self.animationCharacters) == 0:
            return
        self.animationIndex+=1
        value = f'{self.animationCharacters[self.animationIndex%len(self.animationCharacters)]}'
        print(f'[{self.textColor}' + value.ljust(self.size) + f'{LogColors.ENDC}]', end='')

    def animate_bar(self):
        value = min(max(0, self.value), 1)
        fillValue = int(value * self.length)
        emptyValue = self.length - fillValue
        text = ''
        if value > 0:
            text += self.barColors[self.animationIndex%len(self.barColors)]
            text += self.fillBarCharacter * fillValue 
            text += LogColors.ENDC
        if value < 1:
            text += self.emptyBarCharacter * emptyValue
        print(f'[{text}]', end='')

    def animate_percent(self):
        value = int(100 * min(max(0, self.value), 1))
        text = '['
        if(self.value == 1):
            text+=self.percentColor[1]
        else:
            text+=self.percentColor[0]
        text += str(value).rjust(3) + '%' + LogColors.ENDC
        text += ']'
        print(text, end='')
        
    def run(self):
        if self.running:
            return
        self.running = True
        while self.running:
            self.print_progress()
            time.sleep(self.interval)
        else:
            self.print_progress()
    
    def print_progress(self):
        if not self.show:
            return
        self.carriage_return()
        self.animate_text()
        self.animate_bar()
        self.animate_percent()


    def finish(self):
        self.set_message('Done')
        self.set_value(1)
        self.print_progress()
        print()
        self.show = False
        self.set_message('')
        self.set_value(0)

    def __enter__(self):
        self.start()
        return self
    
    def __exit__(self, type, value, traceback):
        self.stop()
    
    def __del__(self):
        self.stop()
        
    def stop(self):
        if self.running:
            self.finish()
            self.running = False
            self.join()
            
def progress(max:int=100, value:int=0, interval:float=0.1) -> Progress:
    return Progress(max, value, interval)

with SharpLock(' '):
    print('sharped locked!')
    
with Progress() as p:
    i = 0
    p.set_message('Testing the threadsnake...')
    while i < 100:
        if random() > 0.5:
            i += 1 
        p.set_value(i / 100.00)
        time.sleep(1)


from threadsnake.core import *
from threadsnake.testing import test, run_test
from typing import Callable, Any
import requests
import time


#set_log_config(LogLevel.ALL)
badge_async('threadsnake 0.0.19', title='version')



port:int = get_port(8088)

print(f'listening on port: {port}')

app = Application(port)

#Configuring top level middlewares
app.configure(static_files('static'))
app.configure(session(Session('threadsnake-session-id')))
app.configure(authorization)
app.configure(body_parser)
app.configure(multipart_form_data_parser('temp', 60))
app.configure(json_body_parser)
app.configure(cors)
app.configure(default_headers(build_default_headers({"purpose":"test"})))
#app.configure(identify_client)
app.configure(header_inspector("content-type", lambda a: print(a)))
app.configure(time_measure)
#app.configure(uses_php('php', 'php', 8086))
app.configure(serve_static_markdown('markdown'))

#Configuring routers
app.use_router(routes_to('routers/users'), '/users')

@app.get('/test-api/base-endpoint/{param}')
def test_api_base_endpoint(app:Application, req:HttpRequest, res:HttpResponse):
    res.end(req.params['param'])

@test({"parameter":"Pelusa"}, True)
def when_param_sent_via_querystring_on_get_it_get_returned_as_response(parameter:str = ''):
    url = f'http://localhost:{port}/test-api/base-endpoint/{parameter}'
    return requests.get(url).text == parameter
    
try:
    app.start()
    time.sleep(1)
    run_test()
except Exception as e:
    print(e)
finally:
    app.stop()


import imp
from .myhttp.http_classes import *
from .myhtml.Html import *
from .pypress_classes import *
from .middlewares import *
from .markdown import *
from .data_access import *
from sys import argv

def get_port(default:int):
    return int(argv[1]) if len(argv) > 1 and argv[1].isdigit() else default



from .myhttp.logging import *
from typing import Callable, Any, Dict

TestFunction = Callable[[Any], bool]
testActions = []

class TestAction:
    def __init__(self, f:Callable, params:Dict[str, Any], expects:bool):
        self.name = f.__name__
        self.action = lambda: f.__call__(**params) == expects

    def __call__(self) -> bool:
        return self.action()

def test(params:Dict[str, Any], expects:bool) -> TestFunction:
    def inner(f:TestFunction) -> TestFunction:
        testActions.append(TestAction(f, params, expects))
        return f
    return inner

def run_test():
    tests = len(testActions)
    success = 0
    while len(testActions) > 0:
        t = testActions.pop()
        badge_async(t.name, LogColors.DARK_BACKGROUND_BLACK, 'testing', LogColors.DARK_BACKGROUND_YELLOW, False)
        if t():
            badge_async(t.name, LogColors.DARK_BACKGROUND_BLACK, 'passed', LogColors.DARK_BACKGROUND_GREEN)
            success += 1
        else:
            badge_async(t.name, LogColors.DARK_BACKGROUND_BLACK, 'failed', LogColors.DARK_BACKGROUND_RED)

    badge_async(str(tests), LogColors.DARK_BACKGROUND_BLACK, 'test cases :', LogColors.DARK_BACKGROUND_BLUE, padding=False)
    badge_async(str(success), LogColors.DARK_BACKGROUND_BLACK, 'test passed:', LogColors.DARK_BACKGROUND_GREEN, padding=False)
    badge_async(str(tests - success), LogColors.DARK_BACKGROUND_BLACK, 'test failed:', LogColors.DARK_BACKGROUND_RED, padding=False)
 


##    threadsnake. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com

from functools import reduce
from typing import Any, Callable, Dict, List, Tuple, Type
from enum import IntEnum
import sqlite3

class FieldType(IntEnum):
    NONE = 0,
    READ = 1,
    WRITE = 2,
    READWRITE = 3,
    ID = 4,
    FILTER = 8

class Field:
    __slots__ = ['name', 'alias', 'type']
    def __init__(self, name:str, alias:str = None, type:FieldType = FieldType.READ) -> None:
        self.name:str = name
        self.alias:str = alias or name
        self.type:FieldType = type
        
    def is_type(self, type:FieldType):
        return self.type & type == type
    
    def apply_type(self, type:FieldType):
        self.type |= type

entities = {}


class BaseEntity:
    def __init__(self, name:str = None):
        self.name:str = name
        self.cls:Type = None
        self.fields:Dict[str, Field] = {}
        
    def with_fields_type(self, fields:List[Any], type:FieldType = FieldType.READWRITE):
        for f in fields:
            alias = f
            if not isinstance(f, str):
                f, alias = f
            else:
                pass
            if f not in self.fields:
                self.fields[f] = Field(f, alias, type)
            else:
                self.fields[f].alias = alias
                self.fields[f].apply_type(type)
        return self
        
    def id(self, fields:Dict[str, str]):
        return self.with_fields_type(fields, FieldType.ID)
    
    def read(self, fields:Dict[str, str]):
        return self.with_fields_type(fields, FieldType.READ)
    
    def write(self, fields:Dict[str, str]):
        return self.with_fields_type(fields, FieldType.WRITE)
    
    def readwrite(self, fields:Dict[str, str]):
        return self.with_fields_type(fields, FieldType.READWRITE)
    
    def filter(self, fields:Dict[str, str]):
        return self.with_fields_type(fields, FieldType.FILTER)
    
class Entity(BaseEntity):
    def __init__(self, name: str = None):
        BaseEntity.__init__(self, name)
        
    def __call__(self, *args: Any, **kwds: Any) -> Any:
        self.cls = args[0]
        self.name = self.name or self.cls.__name__ 
        entities[self.cls] = self
        return args[0]
    
    def fields_with_type(self, type_filter:FieldType):
        return [i for i in self.fields.values() if (i.type & type_filter) == type_filter]

class DatabaseAccess:
    def __init__(self, databaseName:str) -> None:
        self.databaseName = databaseName
        
    def select(self, query:str):
        with sqlite3.connect(self.databaseName) as c:
            cur = c.cursor()
            for r in cur.execute(query):
                yield r
                
    def insert_many(self, tableName:str, columns:List[str], values:List[Tuple[Any]]):
        cols:str = ', '.join(columns)
        par:str = ', '.join(['?' for i in columns])
        query = f'INSERT INTO {tableName} ({cols}) VALUES ({par})'
        try:
            with sqlite3.connect(self.databaseName) as c:
                cur = c.cursor()
                cur.executemany(query, values)
                c.commit()
                return cur.rowcount
        except sqlite3.IntegrityError:
            return 0
        
    def insert(self, tableName:str, values:Dict[str, Any]):
        cols:str = ', '.join([i for i in values.keys()])
        par:str = ', '.join(['?' for i in values.keys()])
        params:Tuple = tuple([values[i] for i in values.keys()])
        query = f'INSERT INTO {tableName} ({cols}) VALUES ({par})'
        try:
            with sqlite3.connect(self.databaseName) as c:
                cur:sqlite3.Cursor = c.cursor()
                cur.execute(query, params)
                c.commit()
                return cur.lastrowid
        except sqlite3.IntegrityError:
            return 0
    
    def select(self, tableName:str, columns:List[str] = None):
        columns = ', '.join(columns or ['*'])
        query:str = f"SELECT {columns} FROM {tableName};"
        with sqlite3.connect(self.databaseName) as c:
            cur = c.cursor()
            for r in cur.execute(query):
                yield r
                        
    def select_match(self, tableName:str, match:Dict[str, Any], columns:List[str] = None):
        columns = ', '.join(columns or ['*'])
        filtering = ' AND '.join([f'{i}=:{i}' for i in match.keys()])
        query:str = f"SELECT {columns} FROM {tableName} WHERE {filtering};"
        with sqlite3.connect(self.databaseName) as c:
            cur = c.cursor()
            for r in cur.execute(query, match):
                yield r

    def select_where(self, tableName:str, where:List[str], columns:List[str] = None):
        columns = ', '.join(columns or ['*'])
        filtering = ' AND '.join(where)
        query:str = f"SELECT {columns} FROM {tableName} WHERE {filtering};"
        with sqlite3.connect(self.databaseName) as c:
            cur = c.cursor()
            for r in cur.execute(query):
                yield r     

    def update(self, tableName, newValues:Dict[str, Any], where:Dict[str, Any]):
        updates = ', '.join([f'{i} = ?' for i in newValues.keys()])
        whereStr = ' AND '.join([f'{i} = ?' for i in where.keys()])
        params = [i for i in newValues.values()] + [i for i in where.values()]
        query:str = f'UPDATE {tableName} SET {updates} WHERE {whereStr};'
        try:
            with sqlite3.connect(self.databaseName) as c:
                cur:sqlite3.Cursor = c.cursor()
                cur.execute(query, params)
                c.commit()
                return cur.rowcount
        except sqlite3.IntegrityError as e:
            return 0
        
    def update_where(self, tableName, newValues:Dict[str, Any], where:List[str]):
        updates = ', '.join([f'{i} = ?' for i in newValues.keys()])
        whereStr = ' AND '.join(where)
        params = [i for i in newValues.values()]
        query:str = f'UPDATE {tableName} SET {updates} WHERE {whereStr};'
        try:
            with sqlite3.connect(self.databaseName) as c:
                cur:sqlite3.Cursor = c.cursor()
                cur.execute(query, params)
                c.commit()
                return cur.rowcount
        except sqlite3.IntegrityError as e:
            return 0
        
    def delete(self, tableName, where:Dict[str, Any]):
        whereStr = ' AND '.join([f'{i} = ?' for i in where.keys()])
        params = [i for i in where.values()]
        query:str = f'DELETE FROM {tableName} WHERE {whereStr};'
        with sqlite3.connect(self.databaseName) as c:
            cur:sqlite3.Cursor = c.cursor()
            cur.execute(query, params)
            c.commit()
            return cur.rowcount
        
    def delete_where(self, tableName, where:List[str]):
        whereStr = ' AND '.join(where)
        query:str = f'DELETE FROM {tableName} WHERE {whereStr};'
        with sqlite3.connect(self.databaseName) as c:
            cur:sqlite3.Cursor = c.cursor()
            cur.execute(query)
            c.commit()
            return cur.rowcount

class NanORM(DatabaseAccess):
    def __init__(self, databaseName:str) -> None:
        super().__init__(databaseName)
    
    def get_entity(self, entity:Any) -> Entity:
        return [i for i in entities.values() if isinstance(entity, i.cls)][0]
    
    def values_of_type(self, entity:Any, type:FieldType) -> Tuple[str, Dict[str, Any]]:
        table, fields = self.fields_of_type(entity, type)
        fields = [i for i in fields if getattr(entity, i.name) is not None]
        return  table, {f.alias:getattr(entity, f.name) for f in fields}
    
    def fields_of_type(self, entity:Any, type:FieldType) -> Tuple[str, Dict[str, Any]]:
        e = self.get_entity(entity)
        return e.name, e.fields_with_type(type)
    
    def entity_insert(self, entity:Any) -> int:
        table, values = self.values_of_type(entity, FieldType.WRITE)
        return self.insert(table, values)

    def entity_delete_by_id(self, entity:Any) -> int:
        table, values = self.values_of_type(entity, FieldType.ID)
        return self.delete(table, values)

    def entity_delete_when_matches(self, entity:Any) -> int:
        table, values = self.values_of_type(entity, FieldType.FILTER)
        return self.delete(table, values)

    def entity_update_by_id(self, entity:Any) -> int:
        table, newValues = self.values_of_type(entity, FieldType.WRITE)
        table, where = self.values_of_type(entity, FieldType.ID )
        return self.update(table, newValues, where)

    def entity_update_when_matches(self, entity:Any) -> int:
        table, newValues = self.values_of_type(entity, FieldType.WRITE)
        table, where = self.values_of_type(entity, FieldType.FILTER)
        return self.update(table, newValues, where)
    
    def entity_select_by_id(self, entity:Any, ctor:Callable[[], Any]) -> int:
        table, where = self.values_of_type(entity, FieldType.ID)
        table, read = self.fields_of_type(entity, FieldType.READ)
        columns = [i.alias for i in read]
        for r in self.select_match(table, where, columns):
            result = ctor()
            for i in range(len(read)):
                setattr(result, read[i].name, r[i])
            yield result
    
    def create_table(self, entity:Any):
        table, read = self.fields_of_type(entity, FieldType.NONE)
        columns = []
        for f in read:
            colName = f.alias
            value = getattr(entity, f.name)
            if isinstance(value, int):
                colName += ' INTEGER'
            elif isinstance(value, float):
                colName += ' FLOAT'
            else:
                colName += ' TEXT'
            if (f.type & FieldType.ID) == FieldType.ID:
                colName += ' PRIMARY KEY'
            columns.append(colName)
        cols = ', '.join(columns)
        query = f'CREATE TABLE IF NOT EXISTS {table}({cols});'
        with sqlite3.connect(self.databaseName) as c:
            cur:sqlite3.Cursor = c.cursor()
            cur.execute(query)
            c.commit()
            return cur.rowcount

@Entity('Person').id([('id', 'rowid')]).readwrite(['firstName', 'lastName', 'identification']).filter(['firstName'])
class PersonExample:
    def __init__(self, id:int = 0, firstName:str = None, lastName:str = None, identification:str = None) -> None:
        self.id = 0
        self.firstName:str = firstName
        self.lastName:str = lastName
        self.identification:str = identification


##    threadsnake. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com

from functools import reduce
import json
from pydoc import cli
import re
import socket
from threading import Thread
import random
import time
from turtle import color
from typing import Any
import uuid
import os
from enum import IntEnum
from .logging import *

class RedirectType(IntEnum):
    MOVED_PERMANENT = 301
    PERMANENT_REDIRECT = 308
    FOUND = 302
    SEE_OTHER = 303
    TEMPORARY_REDIRECT = 307
    MULTIPLE_CHOICE = 301
    NOT_MODIFIED = 304

def get_absolute_path(fileName):
    return os.sep.join([os.sep.join(__file__.split(os.sep)[:-1]), fileName])

responseCodes = {}
threadsnakeIcon = ''

with open(get_absolute_path('response_codes.txt'), 'r') as f:
    for l in [i.strip() for i in f.readlines()]:
        values = l.split('\t')
        if len(values) > 1:
            responseCodes[int(values[0])] = values[1]
        else:
            print(f'Invalid line on response_codes.txt: {values}')

with open(get_absolute_path('icon.txt'), 'rb') as f:
    threadsnakeIcon = f.read().decode('latin1')


contentTypes = {
    "text/": ["html", "css"],
    "text/javascript": ["js"],
    "text/html": ["htm"],
    "application/": ["json", "xml", "pdf", "exe"],
    "image/": ["gif", "png", "jpeg", "bmp", "webp"],
    "image/jpeg": ["jpg"],
    "image/x-icon": ["ico"],
    "audio/": ["mpeg", "webm", "ogg", "midi", "wav"],
    "text/plain": ["txt", "*"]
}

def get_content_type(path:str):
    contentType = 'text/plain'
    log_info('getting content-type from path') ##TOKEN_TO_FIND
    if '.' in path:
        extension:str = path.split(".")[-1:][0]
        for i in contentTypes:
            if extension in contentTypes[i]:
                contentType = i
                if contentType.endswith("/"):
                    contentType += extension
                break
    log_success(f'content-type resolved to {contentType}') ##TOKEN_TO_FIND
    return contentType

def decode_querystring(data:str):
    log_info('decoding querystring') ##TOKEN_TO_FIND
    data = data.replace('+', ' ').replace('%20', ' ')#space
    data = data.replace('%2B', '+')#space
    data = data.replace('%7E', '~')#space
    references = [(i, chr(int(re.findall(r'[\d]+', i)[0]))) for i in re.findall(r'&#[\d]+;', data)]
    for ref in references:
        data = data.replace(ref[0], ref[1])
    log_success(f'querystring resolved') ##TOKEN_TO_FIND
    return data

def map_dictionary(data:str, rowSeparator:str, keySeparator:str):
    return dict([
        tuple([j.strip() for j in i.split(keySeparator, 1)])
        for i in data.split(rowSeparator) if len(i.split(keySeparator)) == 2
    ])

class HttpResponse:
    def __init__(self):
        log_info(f'response created') ##TOKEN_TO_FIND
        self.resposeStatus = 404
        self.resposeText = 'NotFound'
        self.headers = {
            "Content-Type": "text/plain",
            "Content-Length": "0",
            "Connection": "close"
        }
        self.cookieHeaders = []
        self.body = ""
        self.ended = False
        self.contentDisposition = ''
        self.cacheValue = None
        self.encoding = None
        self.lineSep = '\r\n'

    def set_encoding(self, encoding:str):
        log_info(f'encoding set to {encoding}') ##TOKEN_TO_FIND
        self.encoding = encoding
        return self

    def cache(self, cacheValue:str):
        log_info(f'cache set') ##TOKEN_TO_FIND
        self.cacheValue = cacheValue

    def end(self, data:str, status:int = None):
        self.ended = True
        self.status(status or 200)
        log_info(f'response ended') ##TOKEN_TO_FIND
        return self.write(data)

    def redirect(self, url:str, statusCode:int = None):
        self.status(statusCode or RedirectType.TEMPORARY_REDIRECT)
        log_info(f'request redirected to {url}') ##TOKEN_TO_FIND
        self.headers = {"Location": url}
        self.body = ""
        return self

    def status(self, resposeStatus:int, resposeText:str=None):
        self.resposeStatus = resposeStatus
        if resposeText == None and resposeStatus in responseCodes:
            self.resposeText = responseCodes[resposeStatus]
        else:
            self.resposeText = resposeText or "OK"
        return self

    def content_type(self, value:str):
        log_info(f'response content-type set to {value}') ##TOKEN_TO_FIND
        self.headers["Content-Type"] = value
        return self

    def json(self, data:Any):
        self.body = json.dumps(data)
        log_info(f'json response set') ##TOKEN_TO_FIND
        return self.content_type("text/json")

    def write(self, data:str):
        self.body += data
        log_info(f'data with length of {len(data)} bytes append to the body response') ##TOKEN_TO_FIND
        return self

    def html(self, data:str):
        log_info(f'html response set') ##TOKEN_TO_FIND
        return self.content_type('text/html').write(str(data))

    def read_file(self, fileName:str, contentType:str=None):
        with open(fileName, 'r') as f:
            self.body += f.read()
        log_info(f'file {fileName} read from disk') ##TOKEN_TO_FIND
        self.status(200)
        return self.content_type(contentType or get_content_type(fileName))

    def transmit_as_file(self, fileName:str, data:str, contentType:str = None, encoding:str = None):
        self.set_content_disposition('attachment', fileName).set_encoding(encoding or 'UTF-8')
        self.body = data
        log_info(f'data with length of {len(data)} bytes set as file response') ##TOKEN_TO_FIND
        self.status(200)
        return self.content_type(contentType or get_content_type(fileName))

    def set_content_disposition(self, contentDisposition:str, fileName:str = None):
        self.contentDisposition = contentDisposition
        if fileName is not None:
            self.contentDisposition += f'; fileName="{fileName}"'
        log_info(f'content disposition set to {self.contentDisposition}') ##TOKEN_TO_FIND
        return self

    def download_file(self, path:str, fileName:str, contentType:str = None, encoding:str = None):
        self.set_content_disposition('attachment', fileName).set_encoding(encoding or 'UTF-8')
        self.body = ""
        return self.read_file(path, contentType)

    def __str__(self):
        if not self.cacheValue == None:
            return self.cacheValue
        result = f"HTTP/1.1 {self.resposeStatus} {self.resposeText}{self.lineSep}"
        if 'Location' not in self.headers: self.headers["Content-Length"] = len(self.body)
        if len(self.contentDisposition or '') > 0:
            self.headers['Content-Disposition'] = self.contentDisposition 
        mappedHeaders = [f"{i}:{self.headers[i]}{self.lineSep}" for i in self.headers]
        for cookie in self.cookieHeaders:
            print(cookie)
            mappedHeaders.append(f"Set-Cookie:{cookie}{self.lineSep}")
        result += reduce(lambda a, b: a + b, mappedHeaders)
        result += f'{self.lineSep}'
        result += self.body
        log_info(f'response rendered') ##TOKEN_TO_FIND
        return result

    def set_cookie(self, name:str, value:str, durationSec:float = None, domain:str = None, path:str = None):
        cookieString = value
        if durationSec != None: 
            expireUTC = time.strftime("%a, %d %b %Y %H:%M:%S +0000", time.gmtime(time.time() + durationSec))
            cookieString += f"; Expires={expireUTC}"
        if domain != None: cookieString += f"; Domain={domain}"
        if path != None: cookieString += f"; Path={path}"
        log_info(f'cookie set: {name}={cookieString}') ##TOKEN_TO_FIND
        self.cookieHeaders.append(f"{name}={cookieString}")

    def to_bytes(self) -> bytes:
        return str(self).encode(self.encoding or 'latin1')

class HttpRequest:
    __slots__ = [
        'clientAddress', 'raw', 'authorization', 
        'files', 'headers', 'body', 'method', 
        'httpVersion', 'url', 'querystring', 
        'path', 'baseUrl', 'params', 'data', 
        'headers', 'contentType', 'cookies'
    ]
    def __init__(self, raw:str, address:str):
        self.clientAddress = address
        self.load(raw)

    def load(self, raw:str):
        log_load(f'request loaded from {self.clientAddress}') ##TOKEN_TO_FIND
        self.raw = raw
        self.authorization = {}
        self.files = {}
        headerAndBody = self.raw.split('\r\n\r\n', maxsplit=1)
        self.headers = headerAndBody[0].split('\r\n')
        self.body = '' if len(headerAndBody) <= 1 else headerAndBody[1]
        firstLine = self.headers[:1][0].split(' ')
        self.headers = self.headers[1:]
        self.method = firstLine[0]
        self.httpVersion = firstLine[2] if len(firstLine) >= 2 else 'HTTP/1.1'
        self.url = firstLine[1] if len(firstLine) >= 1 else ''
        self.querystring = decode_querystring(self.url.split('?', 1)[1] if '?' in self.url else '')
        self.path = self.url.split('?')[:1][0]
        if not self.path.endswith('/'):
            self.path += '/'
        self.baseUrl = self.url if '?' not in self.url else self.url.split('?', 1)[0]
        self.params = map_dictionary(self.querystring, '&', '=')
        self.data = {}
        self.headers = dict([
            tuple([j.strip() for j in i.split(':', 1)]) for i in self.headers
            if len(i.split(':')) == 2
        ])
        self.contentType = self.headers.get('Content-Type', None)
        self.cookies = {}
        if 'Cookie' in self.headers:
            self.cookies = dict([tuple([j.strip() for j in i.split('=')]) for i in self.headers['Cookie'].split(';') if len(i.split('=')) == 2])

class ServerWorker(Thread):
    def __init__(self, action):
        Thread.__init__(self)
        self.action = action
    
    def run(self):
        self.action()

class Server(Thread):
    def __init__(self, port:int = 80, hostname:str = 'localhost', backlog: int = 5, readTimeout: float = 0.025, bufferSize: int = 1024):
        Thread.__init__(self)
        self.running = False
        self.port = port
        self.hostname = hostname
        self.backlog = backlog
        self.readTimeout = readTimeout
        self.bufferSize = bufferSize
    
    def __enter__(self):
        self.start()
    
    def __exit__(self, type, value, traceback):
        self.stop()
    
    def __del__(self):
        self.stop()
    
    def get_socket(self) -> socket.socket:
        return socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    
    def on_receive(self, data:bytes, clientPort:socket.socket, clientAddress:str):
        raise NotImplementedError()
    
    def on_accept(self, client:socket.socket, address:str, readTimeout:int = None, closeSocket:bool = True):
        self.on_receive(self.read(client, readTimeout), client, address)
        if closeSocket:
            client.close()

    def read(self, client:socket.socket, readTimeout:int = None) -> bytes:
        data = []
        timeout = client.gettimeout()
        try:
            client.settimeout(readTimeout or self.readTimeout)
            while True:
                try:
                    buffer:bytes = client.recv(self.bufferSize)
                    if len(buffer) == 0:
                        break
                    data.extend(buffer)
                except:
                    break
        finally:
            client.settimeout(timeout)
        return bytes(data)
        
    def next_free(self, port:int, max_port=65535):
        sock = self.get_socket()
        while port <= max_port:
            try:
                sock.bind(('', port))
                sock.close()
                return port
            except OSError:
                port += 1
        else:
            raise IOError('no free ports')
    
    def run(self):
        log_warning(f'server starting...') ##TOKEN_TO_FIND
        if self.running:
            return
        self.running = True
        port = self.next_free(self.port)
        if port != self.port:
            log_warning(f'Port {self.port} already in use. Using port {port} instead.')
            self.port = port
        srv = self.get_socket()
        srv.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        srv.bind((self.hostname, self.port))
        srv.listen(self.backlog)
        log_success(f'server sucessfully started') ##TOKEN_TO_FIND
        while self.running:
            try:
                log_success(f'listening for new requests') ##TOKEN_TO_FIND
                (client, address) = srv.accept()
                log_info(f'new requests from {address}') ##TOKEN_TO_FIND
                Thread(target=(lambda :self.on_accept(client, address))).start()
            except Exception as e:
                log_error('unexpected error has occur')
    
    def cancel_listen(self):
        log_warning(f'canceling listen')
        clt = self.get_socket()
        clt.connect((self.hostname, self.port))
        clt.close()
        
    def stop(self):
        if self.running:
            log_warning(f'server stop requested')
            self.running = False
            self.cancel_listen()
            self.join()

class Bridge:
    def __init__(self, port:int, address:str = 'localhost') -> None:
        self.port = port
        self.address = address
        
    def send(self, data:str, encoding:str='latin1', bufferSize:int=1024) -> str:
        result = ''
        target = (self.address, self.port)
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as client:
            client.connect(target)
            client.sendall(data.encode(encoding))
            while data:
                data = client.recv(bufferSize)
                if not data:
                    break
                result += data.decode(encoding)
        return result

class Proxy(Server):
    def __init__(self, port=80, connectionTimeout=None, targets = []):
        Server.__init__(self, port, connectionTimeout)
        self.targets = dict([(i, []) for i in targets])

    def onReceive(self, clientPort:int, data:str, clientAddress):
        target = self.get_target(clientAddress[0])
        result = Bridge(target[1], target[0]).send(data, 'latin1')
        clientPort.send(result.encode('latin1'))
        
    def get_target(self, address:str):
        targets = [i for i in self.targets]
        for t in targets:
            if address in self.targets[t]:
                return t
        t = random.choice(targets)
        self.targets[t].append(address)
        return t

class Session:
    def __init__(self, cookieName = None):
        self.cookieName = cookieName or 'sessionId'
        self.sessions = {}

    def ensure_session(self, req:HttpRequest, res:HttpResponse, reset:bool = False):
        sessionId = ''
        if self.cookieName not in req.cookies:
            sessionId = str(uuid.uuid4())
            res.set_cookie(self.cookieName, sessionId)
            req.cookies[self.cookieName] = sessionId
        else:
            sessionId = req.cookies[self.cookieName]
        if sessionId not in self.sessions or reset:
            self.sessions[sessionId] = {}
        return sessionId

    def get(self, req:HttpRequest, res:HttpResponse, key):
        sessionId = self.ensure_session(req, res)
        if key not in self.sessions[sessionId]:
            return None
        else:
            return self.sessions[sessionId][key]

    def set(self, req:HttpRequest, res:HttpResponse, key, value):
        sessionId = self.ensure_session(req, res)
        self.sessions[sessionId][key] = value

    def create_session(self, req:HttpRequest, res:HttpResponse):
        return RequestSession(req, res, self)

class RequestSession:
    def __init__(self, req:HttpRequest, res:HttpResponse, session:Session):
        self.req = req
        self.res = res
        self.session = session
    
    def __setitem__(self, key:str, value):
        self.session.set(self.req, self.res, key, value)

    def __getitem__(self, key:str):
        return self.session.get(self.req, self.res, key)
    
    def try_get(self, key:str, defaultValue = None):
        value = self.session.get(self.req, self.res, key)
        if value is None:
            value = defaultValue
            self.session.set(self.req, self.res, key, value)
        return value
    
    def reset_session(self):
        self.session.ensure_session(self.req, self.res, True)


##    threadsnake. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com


from enum import IntEnum
import sys
from threading import Lock, Thread, get_ident
from time import strftime
import time
from typing import List, final
import ctypes, ctypes.wintypes
import atexit



if sys.platform == 'win32':
    oldConsoleMode = ctypes.wintypes.DWORD(0)
    def reset_console_mode():
        global oldConsoleMode
        if sys.platform == 'win32':
            kernel = ctypes.windll.kernel32
            kernel.SetConsoleMode(kernel.GetStdHandle(-11), oldConsoleMode)
    kernel = ctypes.windll.kernel32
    kernel.GetConsoleMode(kernel.GetStdHandle(-11), ctypes.byref(oldConsoleMode))
    kernel.SetConsoleMode(kernel.GetStdHandle(-11), 7)
    atexit.register(reset_console_mode)


class LogColorMode(IntEnum):
    NONE = 0
    MESSAGE = 1
    TITLE = 2
    ALL = 3

class LogLevel(IntEnum):
    NONE = 0
    SUCCESS = 1
    INFO = 2
    WARNING = 4
    ERROR = 8
    FATAL = 16
    LOAD = 32
    ALL = 63

class LogInfoTrace(IntEnum):
    NONE = 0
    MOMENT = 1
    THREAD = 2
    
class LogColors:
    BRIGHT_FOREGROUND_WHITE = '\033[97m'
    BRIGHT_FOREGROUND_RED = '\033[91m'
    BRIGHT_FOREGROUND_GREEN = '\033[92m'
    BRIGHT_FOREGROUND_BLUE = '\033[94m'
    BRIGHT_FOREGROUND_YELLOW = '\033[93m'
    BRIGHT_FOREGROUND_MAGENTA = '\033[95m'
    BRIGHT_FOREGROUND_CYAN = '\033[96m'
    BRIGHT_FOREGROUND_BLACK = '\033[90m'
    
    DARK_FOREGROUND_WHITE = '\033[37m'
    DARK_FOREGROUND_RED = '\033[31m'
    DARK_FOREGROUND_GREEN = '\033[32m'
    DARK_FOREGROUND_BLUE = '\033[34m'
    DARK_FOREGROUND_YELLOW = '\033[33m'
    DARK_FOREGROUND_MAGENTA = '\033[35m'
    DARK_FOREGROUND_CYAN = '\033[36m'
    DARK_FOREGROUND_BLACK = '\033[30m'
    
    BRIGHT_BACKGROUND_WHITE = '\033[107m'
    BRIGHT_BACKGROUND_RED = '\033[101m'
    BRIGHT_BACKGROUND_GREEN = '\033[102m'
    BRIGHT_BACKGROUND_BLUE = '\033[104m'
    BRIGHT_BACKGROUND_YELLOW = '\033[103m'
    BRIGHT_BACKGROUND_MAGENTA = '\033[105m'
    BRIGHT_BACKGROUND_CYAN = '\033[106m'
    BRIGHT_BACKGROUND_BLACK = '\033[100m'
    
    DARK_BACKGROUND_WHITE = '\033[47m'
    DARK_BACKGROUND_RED = '\033[41m'
    DARK_BACKGROUND_GREEN = '\033[42m'
    DARK_BACKGROUND_BLUE = '\033[44m'
    DARK_BACKGROUND_YELLOW = '\033[43m'
    DARK_BACKGROUND_MAGENTA = '\033[45m'
    DARK_BACKGROUND_CYAN = '\033[46m'
    DARK_BACKGROUND_BLACK = '\033[40m'
    ENDC = '\033[0m'

colors = {
    LogLevel.SUCCESS : LogColors.BRIGHT_FOREGROUND_GREEN,
    LogLevel.INFO : LogColors.BRIGHT_FOREGROUND_BLUE,
    LogLevel.WARNING : LogColors.BRIGHT_FOREGROUND_YELLOW,
    LogLevel.ERROR: LogColors.BRIGHT_FOREGROUND_RED,
    LogLevel.FATAL: LogColors.DARK_BACKGROUND_RED + LogColors.DARK_FOREGROUND_BLACK,
    LogLevel.LOAD: LogColors.BRIGHT_FOREGROUND_MAGENTA
}

prefixes = {
    LogLevel.SUCCESS : '[done]',
    LogLevel.INFO : '[info]',
    LogLevel.WARNING : '[warn]',
    LogLevel.ERROR: '[erro]',
    LogLevel.FATAL: '[fatl]',
    LogLevel.LOAD: '[load]' 
}

logLevel:LogLevel = LogLevel.NONE
logMode:LogColorMode = LogColorMode.TITLE
logInfoTrace:LogInfoTrace = LogInfoTrace.MOMENT | LogInfoTrace.THREAD
logLock:Lock = Lock()

def set_log_config(level:LogLevel = LogLevel.NONE, mode:LogColorMode = LogColorMode.TITLE, infoTrace:LogInfoTrace = LogInfoTrace.MOMENT):
    global logLevel, logMode, logInfoTrace
    logLevel = level
    logMode = mode
    logInfoTrace = infoTrace

def toggle_log_level(level:LogLevel):
    global logLevel
    logLevel ^= level

def log(message:str, title:str = None, level:LogLevel = LogLevel.INFO, threadId:int = 0):
    global logLevel, logMode, colors, prefixes, logLock
    try:
        logLock.acquire()
        if level & logLevel == level and level in colors:
            color = colors[level]
            logPrefix = '>' if level not in prefixes else prefixes[level]
            title = logPrefix if title is None else f'{logPrefix} : {title}'
            if (logInfoTrace & LogInfoTrace.THREAD) == LogInfoTrace.THREAD:
                title = f'(thread:{str(threadId).zfill(7)}) ' + title
            if (logInfoTrace & LogInfoTrace.MOMENT) == LogInfoTrace.MOMENT:
                title = strftime('%Y-%m-%d %H:%M:%S') + ' ' + title
            fullMessage = title if (logMode & LogColorMode.TITLE != LogColorMode.TITLE) else color + title + LogColors.ENDC
            fullMessage += ' '
            fullMessage += message if (logMode & LogColorMode.MESSAGE != LogColorMode.MESSAGE) else color + title + LogColors.ENDC
            print(fullMessage)
    except Exception as e:
        print(e)
    finally:
        logLock.release()

def badge(text:str, textBackground:LogColors = LogColors.DARK_BACKGROUND_BLUE, title:str = None, titleBackground:LogColors = LogColors.DARK_BACKGROUND_BLACK, padding:bool = True):
    message = titleBackground + f' {title or ""} ' + textBackground + f' {text} ' + LogColors.ENDC
    try:
        logLock.acquire()
        if padding:
            print()
        print(message)
        if padding:
            print()
    except Exception as e:
        print(e)
    finally:
        logLock.release()

def badge_async(text:str, textBackground:LogColors = LogColors.DARK_BACKGROUND_BLUE, title:str = None, titleBackground:LogColors = LogColors.DARK_BACKGROUND_BLACK, padding:bool = True):
    Thread(target=lambda: badge(text, textBackground, title, titleBackground, padding)).start()
        
def log_async(message:str, title:str = None, level:LogLevel = LogLevel.INFO):
    Thread(target=lambda: log(message, title, level, get_ident())).start()

def log_success(message:str, title:str = None) -> None:
    log_async(message, title, LogLevel.SUCCESS)
    
def log_info(message:str, title:str = None) -> None:
    log_async(message, title, LogLevel.INFO)
    
def log_error(message:str, title:str = None) -> None:
    log_async(message, title, LogLevel.ERROR)
    
def log_warning(message:str, title:str = None) -> None:
    log_async(message, title, LogLevel.WARNING)
    
def log_fatal(message:str, title:str = None) -> None:
    log_async(message, title, LogLevel.FATAL)
    
def log_load(message:str, title:str = None) -> None:
    log_async(message, title, LogLevel.LOAD)
    

class Progress(Thread):
    def __init__(self, length:int=25, value:int=0, interval:float=0.1, emptyCharacter:str = '.', fillCharacter:str = '|'):
        Thread.__init__(self)
        
        self.running:bool = False
        self.length:int = length
        self.value:float = value
        self.interval:float = interval
        
        self.textColor:str = LogColors.BRIGHT_FOREGROUND_WHITE
        self.barColors:List[str] = [LogColors.BRIGHT_FOREGROUND_GREEN]
        self.percentColor:str = [LogColors.BRIGHT_FOREGROUND_YELLOW, LogColors.BRIGHT_FOREGROUND_GREEN]
        
        
        self.animationCharacters:List[str] = [' ']
        self.animationIndex = 0
        self.size = 10
        self.emptyBarCharacter = ' '
        self.fillBarCharacter = '█'
        
        self.show:bool = True
        
    def set_message(self, text:str):
        text = ' '*self.size + text + ' '*self.size
        self.animationCharacters:List[str] = [text[:self.size] for i in range(10)] 
        self.animationCharacters += [text[i:i+self.size] for i in range(len(text)-self.size)]
    
    def set_value(self, value:float = 0):
        self.value = min(max(0, value), 1)
        self.show = True
    
    def carriage_return(self):
        print('\r', end='')
    
    def animate_text(self):
        if len(self.animationCharacters) == 0:
            return
        self.animationIndex+=1
        value = f'{self.animationCharacters[self.animationIndex%len(self.animationCharacters)]}'
        print(f'[{self.textColor}' + value.ljust(self.size) + f'{LogColors.ENDC}]', end='')

    def animate_bar(self):
        value = min(max(0, self.value), 1)
        fillValue = int(value * self.length)
        emptyValue = self.length - fillValue
        text = ''
        if value > 0:
            text += self.barColors[self.animationIndex%len(self.barColors)]
            text += self.fillBarCharacter * fillValue 
            text += LogColors.ENDC
        if value < 1:
            text += self.emptyBarCharacter * emptyValue
        print(f'[{text}]', end='')

    def animate_percent(self):
        value = int(100 * min(max(0, self.value), 1))
        text = '['
        if(self.value == 1):
            text+=self.percentColor[1]
        else:
            text+=self.percentColor[0]
        text += str(value).rjust(3) + '%' + LogColors.ENDC
        text += ']'
        print(text, end='')
        
    def run(self):
        if self.running:
            return
        self.running = True
        while self.running:
            self.print_progress()
            time.sleep(self.interval)
        else:
            self.print_progress()
    
    def print_progress(self):
        if not self.show:
            return
        self.carriage_return()
        self.animate_text()
        self.animate_bar()
        self.animate_percent()


    def finish(self):
        self.set_message('Done')
        self.set_value(1)
        self.print_progress()
        print()
        self.show = False
        self.set_message('')
        self.set_value(0)

    def __enter__(self):
        self.start()
        return self
    
    def __exit__(self, type, value, traceback):
        self.stop()
    
    def __del__(self):
        self.stop()
        
    def stop(self):
        if self.running:
            self.finish()
            self.running = False
            self.join()
  


##    threadsnake. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com

import base64
import json
import os
import re
from hashlib import md5
from typing import Dict

def get_absolute_path(fileName):
    return os.sep.join([os.sep.join(__file__.split(os.sep)[:-1]), fileName])

def encode_base_64(data:str, encoding:str = 'latin1'):
    return base64.b64encode(data.encode(encoding)).decode(encoding)

def decode_base_64(data:str, encoding:str = 'latin1'):
    return base64.b64decode(data.encode(encoding)).decode(encoding)

def load_data(markdownJsonName:str):
    filename = os.sep.join(['markdown_data', markdownJsonName])
    filename = get_absolute_path(filename)
    with open(filename, 'r', encoding='latin1') as f:
        return json.loads(f.read())



commonExpressions = load_data('general.txt')#json.loads(decode_base_64('ewogICAgIiZndDsgKFtcXHdcXFddKj8pKFxcclxcblxcclxcbnxcXHJcXHJ8XFxuXFxuKSI6ICI8YmxvY2txdW90ZT5cXDE8L2Jsb2NrcXVvdGU+XFwyIiwKICAgICIoXFxyXFxufFxccnxcXG4pezIsfSI6ICJcclxuPGJyLz5cclxuIiwKICAgICJfKFteX10rPylfIjogIjxpPlxcMTwvaT4iLAogICAgIlxcKlxcKihbXFx3XFxXXSs/KVxcKlxcKiI6ICI8Yj5cXDE8L2I+IiwKICAgICIjIyMjIyNbXFxzXSooW15cXHJcXG5dKykiOiAiPGg2PlxcMTwvaDY+IiwKICAgICIjIyMjI1tcXHNdKihbXlxcclxcbl0rKSI6ICI8aDU+XFwxPC9oNT4iLAogICAgIiMjIyNbXFxzXSooW15cXHJcXG5dKykiOiAiPGg0PlxcMTwvaDQ+IiwKICAgICIjIyNbXFxzXSooW15cXHJcXG5dKykiOiAiPGgzPlxcMTwvaDM+IiwKICAgICIjI1tcXHNdKihbXlxcclxcbl0rKSI6ICI8aDI+XFwxPC9oMj4iLAogICAgIiNbXFxzXSooW15cXHJcXG5dKykiOiAiPGgxPlxcMTwvaDE+PGhyLz4iLAogICAgIiFcXFsoW15cXF1dKz8pXFxdXFxzKlxcKChbXlxcKV0rPylcXCkiOiAiPGltZyBocmVmPVwiXFwyXCIgYWx0PVwiXFwxXCIvPiIsCiAgICAiXFxbKFteXFxdXSs/KVxcXVxccypcXCgoW15cXCldKz8pXFwpIjogIjxhIGhyZWY9XCJcXDJcIj5cXDE8L2E+IiwKICAgICJgYGAoW1xcd1xcV10rPylgYGAiOiAiPGRpdiBzdHlsZT1cImJhY2tncm91bmQtY29sb3I6IzAwMDc7IHBhZGRpbmc6MC4zZW0gMmVtO1wiPlxcMTwvZGl2PiIsCiAgICAiKChbXFxkXStcXC5bXlxcclxcbl0rW1xcclxcbl0rKSspIjogewogICAgICAgICJyZXAiOiAiPG9sPlxcMTwvb2w+IiwKICAgICAgICAiaW5uZXIiOiB7CiAgICAgICAgICAgICJbXFxkXStcXC4oW15cXHJcXG5dKykiOiAiPGxpPlxcMTwvbGk+IgogICAgICAgIH0KICAgIH0sCiAgICAiKChcXCpbXlxcclxcbl0rW1xcclxcbl0rKSspIjogewogICAgICAgICJyZXAiOiAiPHVsPlxcMTwvdWw+IiwKICAgICAgICAiaW5uZXIiOiB7CiAgICAgICAgICAgICJcXCooW15cXHJcXG5dKykiOiAiPGxpPlxcMTwvbGk+IgogICAgICAgIH0KICAgIH0KfQ=='))
pythonExpressions = load_data('python.txt')#json.loads(decode_base_64('ewogICAgImBgYHB5dGhvbihbXFx3XFxXXSs/KWBgYCI6IHsKICAgICAgICAicmVwIjogIjxkaXYgc3R5bGU9XCJwYWRkaW5nOjBlbSAxZW07IGNvbG9yOiNCQ0Y7IGJhY2tncm91bmQ6YmxhY2s7IGZvbnQtZmFtaWx5OmNhbmRhcmE7XCI+XFwxPC9kaXY+IiwKICAgICAgICAiaW5uZXIiOiB7CiAgICAgICAgICAgICIoXFxyXFxufFxccnxcXG4pIjogIlxyXG48YnIvPlxyXG4iLAogICAgICAgICAgICAiKCAgICB8XHQpIjogIiZuYnNwOyZuYnNwOyZuYnNwOyZuYnNwOyIsCiAgICAgICAgICAgICIoJycnW1xcd1xcV10qPycnJ3wnW1xcd1xcV10qPycpIjogewogICAgICAgICAgICAgICAgInJlcCI6ICI8c3BhbiBzdHlsZT1cImNvbG9yOiNGOTQ7XCI+XFwxPC9zcGFuPiIsCiAgICAgICAgICAgICAgICAiaW5uZXIiOiB7CiAgICAgICAgICAgICAgICAgICAgIihbe31dKSI6ICI8c3BhbiBzdHlsZT1cImNvbG9yOmJsdWU7XCI+XFwxPC9zcGFuPiIKICAgICAgICAgICAgICAgIH0KICAgICAgICAgICAgfSwKICAgICAgICAgICAgIihbXFx3XFxkX10rKVxcKCI6ICI8c3BhbiBzdHlsZT1cImNvbG9yOiNGRjY7XCI+XFwxPC9zcGFuPigiLAogICAgICAgICAgICAiKFtcXFddKShhc3xhc3NlcnR8YnJlYWt8Y29udGludWV8ZGVsfGVsaWZ8ZWxzZXxleGNlcHR8ZmluYWxseXxmb3J8ZnJvbXxpZnxpbXBvcnR8cGFzc3xyYWlzZXxyZXR1cm58dHJ5fHdoaWxlfHdpdGh8eWllbGQpKFtcXFddKSI6ICJcXDE8c3BhbiBzdHlsZT1cImNvbG9yOiNGNEY7XCI+XFwyPC9zcGFuPlxcMyIsCiAgICAgICAgICAgICIoW1xcV10pKGFuZHxjbGFzc3xkZWZ8RmFsc2V8Z2xvYmFsfGlufGlzfGxhbWJkYXxOb25lfG5vbmxvY2FsfG5vdHxvcnxUcnVlKShbXFxXXSkiOiAiXFwxPHNwYW4gc3R5bGU9XCJjb2xvcjojNjZGO1wiPlxcMjwvc3Bhbj5cXDMiLAogICAgICAgICAgICAiKFt7fVxcW1xcXVxcKFxcKV0rKSI6ICI8c3BhbiBzdHlsZT1cImNvbG9yOndoaXRlO1wiPlxcMTwvc3Bhbj4iLAogICAgICAgICAgICAiKFxcQCkiOiAiPHNwYW4gc3R5bGU9XCJjb2xvcjpncmVlbjtcIj5cXDE8L3NwYW4+IgogICAgICAgIH0KICAgIH0KfQ=='))

replacements = {
    '0':'B', '1':'C', '2':'D', '3':'F',
    '4':'G', '5':'H', '6':'J', '7':'K',
    '8':'L', '9':'M', 'a':'N', 'b':'P',
    'c':'Q', 'd':'R', 'e':'S', 'f':'T',
    }
def get_hash(content:str):
    res = md5(content.encode('latin1')).hexdigest()
    for r in replacements:
        res = res.replace(r, replacements[r]) 
    return res

class MarkdownToHtmlParser:
    def __init__(self) -> None:
        self.expressions = {i:i for i in []}

    def use(self, expressionSet:Dict[str, str]):
        self.expressions.update(expressionSet)
        return self
    
    def use_common(self):
        return self.use(commonExpressions)

    def use_python(self):
        return self.use(pythonExpressions)

    def apply_regex_set(self, data:str, regexset:Dict):
        replacements = {}
        styles = {}
        for r in regexset:
            rep = regexset[r]
            if isinstance(rep, str):
                data = re.sub(r, rep, data)
                #while re.search(r, data):
                #    match = re.search(r, data).group()
                #    match = re.sub(r, rep, match, 1)
                #    hash = get_hash(match)
                #    replacements[hash] = match
                #    data = re.sub(r, hash, data, 1)
            elif 'inner' in rep and 'rep' in rep:
                if 'styles' in rep:
                    styles[get_hash(rep['styles'])] = rep['styles']
                while re.search(r, data):
                    match = re.search(r, data).group()
                    match = re.sub(r, rep['rep'], match, 1)
                    match, reps, styl = self.apply_regex_set(match, rep['inner'])
                    replacements.update(reps)
                    styles.update(styl)
                    hash = get_hash(match)
                    replacements[hash] = match
                    data = re.sub(r, hash, data, 1)
        return data, replacements, styles
    
    def convert(self, fileName:str, output:str):
        def tag(name:str, content:str):
            return f'<{name}>{content}</{name}>'
        data, style = self.parse_file(fileName)
        html = tag('html', tag('head', tag('style', style)) + tag('body', data))
        with open(output, 'w', encoding='latin1') as w:
            w.write(html)
    
    def parse_file(self, fileName:str):
        with open(fileName, 'r', encoding='latin1') as r:
            return self.parse_str(r.read())
            
    def parse_str(self, data:str):
        data = data.replace('>','&gt;').replace('<', '&lt;')
        data, replacements, styles = self.apply_regex_set(data, self.expressions)
        while len([i for i in replacements if i in data]) > 0:
            for r in replacements:
                data = data.replace(r, replacements[r])
        stylesString = '\r\n'.join([styles[i] for i in styles])
        return data, stylesString

def create_python_markdown_parser():
    return MarkdownToHtmlParser().use_python().use_common()



##    threadsnake. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com

from functools import reduce
from typing import Dict, List, Tuple

class HtmlNode:
    def render(self, tabLevel:int = 0):
        pass

class HtmlText(HtmlNode):
    def __init__(self, content:str = ''):
        super().__init__()
        self.content = content
    
    def render(self, tabLevel: int = 0):
        return '\t'*tabLevel + f'{self.content}'

class HtmlElement(HtmlNode):
    def __init__(self, tagName:str, attributes:Dict[str, str] = None, children:List[HtmlNode] = None):
        self.tagName = tagName
        self.attributes = attributes or dict()
        self.children = children or []
        
    def __getitem__(self, index):
        res = []
        for i in self.children:
            if not isinstance(i, HtmlElement):
                print('nothtml')
                continue
            if isinstance(index, int) and len(self.children) > index:
                res = [self.children[index]]
                break
            if not isinstance(index, str):
                raise Exception("Name should be int or string")
            if index.startswith('.') and 'class' in i.attributes and index[1:] in i.attributes['class'].split(' '):
                res.append(i)
            elif index.startswith('#') and 'id' in i.attributes and i.attributes['id'] == index[1:]:
                res.append(i)
            elif i.tagName == index:
                res.append(i)
        return res
    
    def for_each(self, selector, action):
        for c in self[selector]:
            action(c)
    
    def render(self, tabLevel:int = 0):
        result = tabLevel * '\t' + f'<{self.tagName}'
        if(len(self.attributes) == 0 and len(self.children) == 0):
            result += '/>\n'
        else:
            if len(self.attributes) > 0:
                result += reduce(lambda a, b: a+b, [f' {i}="{self.attributes[i]}"' for i in self.attributes])
            result += '>\n'
            for c in self.children:
                result += c.render(tabLevel + 1)
            result += tabLevel * '\t' + f'</{self.tagName}>\n'
        return result
    
    def append_child(self, *children:HtmlNode):
        for child in children:
            self.children.append(child)
        return self

    def set_attribute(self, attributeName:str, attributeValue:str):
        self.attributes[attributeName] = attributeValue
        return self
    
def script(src:str):
    return HtmlElement('script', {'src':src})
    
def style(href:str):
    return HtmlElement('link', {'rel':'stylesheet','href':href})
    
def head(title:str, scripts:List[str] = [], styles:List[str] = []):
    result = HtmlElement('head', children=[text_element(title, 'title')])
    for stl in styles:
        result.append_child(style(stl))
    for scr in scripts:
        result.append_child(script(scr))
    return result

def link(text:str, href:str):
    return HtmlElement('a', {'href':href}).append_child(HtmlText(text))

def text_element(text:str, tag:str = 'div', attributes:Dict[str, str] = None):
    return HtmlElement(tag, attributes or dict()).append_child(HtmlText(text))

def html(title:str):
    result = HtmlElement('html').append_child(head(title))
    result.append_child(HtmlElement('body', children=[link('google', 'www.google.com')]))
    return result

def page(title:str, scripts:List[str] = [], styles:List[str] = []) -> Tuple[HtmlElement, HtmlElement]:
    content = HtmlElement('div', {"class":"mainContent"})
    html = HtmlElement('html').append_child(head(title, scripts, styles))
    html.append_child(HtmlElement('body', children=[content]))
    return html, content

def header(title:str):
    return text_element(title, 'header')

def nodes_group(tag:str, *children:HtmlNode):
    result = HtmlElement(tag)
    for c in children:
        result.append_child(c)
    return result

def links_group(tag:str, links:Dict[str, str] = dict()):
    result = HtmlElement(tag)
    for l in links:
        result.append_child(link(l, links[l]))
    return result

def nav(links:Dict[str, str] = dict()):
    return links_group('nav', links)

def aside(links:Dict[str, str] = dict()):
    return links_group('aside', links)


##    threadsnake. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com


from functools import reduce
import socket
import time
from types import FunctionType, ModuleType
from typing import Any, Callable, Dict, List, Union

from threadsnake.myhttp.logging import log_error, log_info, log_load, log_success, log_warning

from .myhttp import Server, HttpRequest, HttpResponse, Session, RequestSession
from uuid import uuid4
import os
import re
from importlib import __import__ as importpy
import importlib.util



'''
Loads a library from specified path.
Thanks to 
[[https://stackoverflow.com/users/7779/sebastian-rittau|Sebastian Rittau|target="_blank"]] and 
[[https://stackoverflow.com/users/3907364/raven|Raven|target="_blank"]]
'''
def import_library(name, path, modulePreload:Callable[[ModuleType],None] = None):
    spec = importlib.util.spec_from_file_location(name, path)
    mod = importlib.util.module_from_spec(spec)
    if modulePreload is not None:
        modulePreload(mod)
    spec.loader.exec_module(mod)
    return mod

class RouterTemplate:
    def __init__(self) -> None:
        self.R:Router = None

class Router:
    def __init__(self):
        #print('router created')
        self.routes = {}
        self.callbackMutator:CallbackMutator = lambda a: a
        self.logRouteConfig = False

    def register_function(self, httpMethod:str, route:str):
        ref:Router = self
        if not route.endswith('/'):
            route += '/'
        if not route.startswith('/'):
            route = '/' + route
        while '//' in route:
            route = route.replace('//', '/')
        if self.logRouteConfig:
            log_warning(f"Route configured [{httpMethod.upper()}] {route}") ##TOKEN TO FIND
        def inner(function:Callback) -> Callback:
            function = self.callbackMutator(function)
            if httpMethod.upper() not in ref.routes:
                ref.routes[httpMethod.upper()] = {route: function}
            else:
                ref.routes[httpMethod.upper()][route] = function
            return function
        return inner
    
    def use_globally(self, mutator):
        callbackMutator = self.callbackMutator
        def innerMutator(c):
            return mutator(callbackMutator(c))
        self.callbackMutator = innerMutator
        return self

    def get(self, route):
        return self.register_function('GET', route)

    def post(self, route):
        return self.register_function('POST', route)
    
    def put(self, route):
        return self.register_function('PUT', route)
    
    def delete(self, route):
        return self.register_function('DELETE', route)

    def configure(self, handler):
        self.stack.append(handler)
        return self

    def use_router(self, router, root):
        for method in router.routes:
            for action in router.routes[method]:
                self.register_function(method,f'{root}{action}')(router.routes[method][action])
        return self

    def serve_statically(self, route:str, content:Union[str, Callable[[Any, HttpRequest, HttpResponse],str]], encoding:str=None):
        def sub(s:str, ss:str)->str:
            return s if ss not in s else s[s.index(ss)+1:]
        fileName = sub(sub(route, '/'), '\\')
        def callback(app:Application, req:HttpRequest, res:HttpResponse):
            data = content if isinstance(content, str) else content(app, req, res)
            res.transmit_as_file(fileName, data, encoding=encoding)
        self.get(route)(callback)
        return self

    def __getattr__(self, method):
        def inner(route):
            return self.register_function(method.upper(), route)
        return inner


class Application(Server, Router):
    def __init__(self, port: int = 80, hostname: str = 'localhost', backlog: int = 5, readTimeout: float = 0.1, bufferSize: int = 1024):
        Server.__init__(self, port, hostname, backlog, readTimeout, bufferSize)
        Router.__init__(self)
        self.stack = []
        self.session:RequestSession = None
        self.logRouteConfig = True
        
    def cli_loop(self, main_event: Callable):
        self.start()
        continueLoop = True
        while(continueLoop):
            try:
                continueLoop = main_event()
            except KeyboardInterrupt as e:
                print('[CLI EXIT]')
                break
            finally:
                self.stop()
                
    def wait_exit(self, message:str = 'press [Enter] to exit...'):
        try:
            self.start()
            log_info(message)
            input('')
        except:
            pass
        finally:
            self.stop()

    def on_receive(self, data:bytes, clientPort:socket.socket, clientAddress):
        data_s = data.decode('latin1')
        if(len(data_s) == 0): return
        req = None
        res = HttpResponse()
        try:
            req = HttpRequest(data_s, clientAddress)
        except Exception as e:
            log_error(e) ##TOKEN TO FIND
            res.status(403, "BadRequest")
            clientPort.send(str(res).encode('latin1'))
            return
        
        if req.headers.get('Connection', '').lower() == 'keep-alive':
            log_warning(f'Connection from {clientAddress} requested kept-alive!')
            chunk:bytes = self.read(clientPort, 0.2 if len(data_s) < 128 else 1)
            if len(chunk) != 0:
                self.on_receive(b''.join([data, chunk]), clientPort, clientAddress)
                return
            else:
                log_success(f'{clientAddress} ended...')

        log_info(f'response pipeline created') ##TOKEN TO FIND
        stack = self.stack.copy()

        def next():
            if len(stack) > 0 and not res.ended:
                stack.pop()(self, req, res, next)
        
        if ':' in req.path:
            log_warning(f'potential query:pass params detected') ##TOKEN TO FIND
            pathParts = req.path.split('/')
            newPath = []
            for i in pathParts:
                queryPassParam = i.split(':')
                if len(queryPassParam) == 2:
                    log_load(f'query:pass param {queryPassParam[0]} resolved to {queryPassParam[1]}') ##TOKEN TO FIND
                    req.params[queryPassParam[0]] = queryPassParam[1]
                else:
                    newPath.append(i)
            req.path = '/'.join(newPath)

        if req.method in self.routes:
            log_success(f'method {req.method} found in registered routes') ##TOKEN TO FIND
            regularPaths = [i for i in self.routes[req.method]]
            for route in regularPaths:
                pattern = route
                pattern = re.sub(r"{([\w]+)\:int}", r"(?P<\1>[-]?[\\d]+)", pattern)
                pattern = re.sub(r"{([\w]+)\:float}", r"(?P<\1>[-]?[\\d]+[\\.]?[\\d]?)", pattern)
                pattern = re.sub(r"{([\w]+)\:re\(([\w\W]+?)\)}", r"(?P<\1>\2)", pattern)
                pattern = re.sub(r"{([\w]+)}", r"(?P<\1>[\\w]+)", pattern)
                pattern = "^" + pattern + "$"
                match = re.match(pattern, req.path)
                if match:
                    log_success(f'request {req.url} matches {route}') ##TOKEN TO FIND
                    handler = self.routes[req.method][route]
                    queryParams = match.groupdict()
                    for i in queryParams:
                        log_load(f'param {i} set to {queryParams[i]}') ##TOKEN TO FIND
                        req.params[i] = queryParams[i]
                    log_success(f'middleware for {route} add to pipeline') ##TOKEN TO FIND
                    def middleware(app, req, res, next):
                        res.status(200)
                        handler(app, req, res)
                        next()
                    stack.append(middleware)
                    break
        stack.reverse()
        log_info(f'pipeline begin') ##TOKEN TO FIND
        next()
        log_info(f'pipeline end') ##TOKEN TO FIND
        clientPort.send(res.to_bytes())

Middleware = Callable[[Application, HttpRequest, HttpResponse, Callable[[], None]], None]
Callback = Callable[[Application, HttpRequest, HttpResponse], None]
ServerCallback = Callable[[HttpRequest, HttpResponse], None]
DictProvider = Callable[[], Dict[str, Any]]
CallbackMutator = Callable[[Callback], Callback]

def build_application(port: int, middlewares: List[Middleware] = None, routers: Dict[str, Router] = None):
    app = Application(port)
    for m in middlewares or []:
        app.configure(m)
    for r in routers or dict():
        app.use_router(routers[r], r)
    return app

def create_server(port:int, callback:ServerCallback) -> Application:
    app = Application(port)
    def innerMiddleware(application:Application, req:HttpRequest, res:HttpResponse, next):
        callback(req, res)
        next()
    app.configure(innerMiddleware)
    return app


routers:Dict[str, ModuleType] = {}
def load_router(name, path) -> RouterTemplate:
    def assign_router(mod:ModuleType) -> None:
        global routers
        mod.R = Router()
        routers[name] = mod
    return import_library(name, path, assign_router)

def export(caller:str) -> Router:
    global routers
    return routers[caller].R
    
def routes_to(name:str):
    path = name if name.endswith('.py') else name + '.py'
    libName = 'threadsnake.router.to_'+str(uuid4()).replace('-', '')
    return load_router(libName, path).R

def routes_to_folder(folder:str = 'routers', onExcept:Callable[[Exception,], None] = None) -> Dict[str, Router]:
    result:Dict[str, Router] = {}
    files = os.listdir(folder)
    for f in [(i[:-3], os.sep.join(folder, i)) for i in files if i.endswith('.py')]:
        try:
            result = result[f[0]] = load_router(f[0], f[1]).R
        except Exception as e:
            if onExcept is not None:
                onExcept(e)
    return result
